"""
$(TYPEDEF)

$(TYPEDFIELDS)

This parametric type allows to define a new dot product from the one saved in `dt::dot`. More precisely:

	dt(u1, u2, p1::T, p2::T, theta::T) where {T <: Real}

computes, the weigthed dot product ``\\langle (u_1,p_1), (u_2,p_2)\\rangle_\\theta = \\theta \\Re \\langle u_1,u_2\\rangle  +(1-\\theta)p_1p_2`` where ``u_i\\in\\mathbb R^N``. The ``\\Re`` factor is put to ensure a real valued result despite possible complex valued arguments.

!!! info "Info"
    This is used in the pseudo-arclength constraint with the dot product ``\\frac{1}{N} \\langle u_1,u_2\\rangle,\\quad u_i\\in\\mathbb R^N``
"""
struct DotTheta{Tdot, Ta}
	"dot product used in pseudo-arclength constraint"
	dot::Tdot
	"Linear operator associated with dot product, i.e. dot(x, y) = <x, Ay>, where <,> is the standard dot product on R^N. You must provide an inplace function which evaluates A. For example `x -> rmul!(x, 1/length(x))`."
	apply!::Ta
end

DotTheta() = DotTheta( (x, y) -> dot(x, y) / length(x), x -> rmul!(x, 1/length(x)))
DotTheta(dt) = DotTheta(dt, nothing)

# we restrict the type of the parameters because for complex problems, we still want the parameter to be real
(dt::DotTheta)(u1, u2, p1::T, p2::T, θ::T) where {T <: Real} = real(dt.dot(u1, u2) * θ + p1 * p2 * (one(T) - θ))

# Implementation of the norm associated to DotTheta
# we restrict the type of the parameters because for complex problems, we still want the parameter to be real
(dt::DotTheta)(u, p::T, θ::T) where T = sqrt(dt(u, u, p, p, θ))

(dt::DotTheta)(a::BorderedArray{vec, T}, b::BorderedArray{vec, T}, θ::T) where {vec, T} = dt(a.u, b.u, a.p, b.p, θ)
(dt::DotTheta)(a::BorderedArray{vec, T}, θ::T) where {vec, T} = dt(a.u, a.p, θ)
####################################################################################################
# equation of the arc length constraint
arcLengthEq(dt::DotTheta, u, p, du, dp, θ, ds) = dt(u, du, p, dp, θ) - ds
####################################################################################################
abstract type AbstractTangentPredictor end
abstract type AbstractSecantPredictor <: AbstractTangentPredictor end

getPredictor!(iter::AbstractContinuationIterable, state::AbstractContinuationState, nrm = false) = getPredictor!(iter, state, iter.tangentAlgo, nrm)

# reset the predictor
Base.empty!(::Union{Nothing, AbstractTangentPredictor}) = nothing

# this function only mutates z_pred
# the nrm argument allows to just increment z_pred.p by ds
function getPredictor!(iter::AbstractContinuationIterable,
					state::AbstractContinuationState,
					pred::Talgo,
					nrm = false) where {Talgo <: AbstractTangentPredictor}
	# we perform z_pred = z + ds * τ
	copyto!(state.z_pred, state.z)
	ds = state.ds
	nrm ? axpy!(ds / state.τ.p, state.τ, state.z_pred) : axpy!(ds, state.τ, state.z_pred)
end

# generic corrector based on Bordered formulation
function corrector(it, state, pred::AbstractTangentPredictor;
			normC = norm, callback = cbDefault, kwargs...)
	if state.z_pred.p <= it.contParams.pMin || state.z_pred.p >= it.contParams.pMax
		state.z_pred.p = clampPredp(state.z_pred.p, it)
		return corrector(it, state, NaturalPred(); normC = normC, callback = callback, kwargs...)
	end
	return newtonPALC(it, state; linearbdalgo = it.linearAlgo, normN = normC, callback = callback, kwargs...)
end
####################################################################################################
"""
	Natural predictor / corrector
"""
struct NaturalPred <: AbstractTangentPredictor end

function getPredictor!(iter::AbstractContinuationIterable,
						state::AbstractContinuationState,
						pred::NaturalPred, nrm = false)
	copyto!(state.z_pred, state.z)
	state.z_pred.p += state.ds
end

# corrector based on natural formulation
function corrector(it, state, pred::NaturalPred; normC = norm, callback = cbDefault, kwargs...)
	res = newton(it.F, it.J, state.z_pred.u, setParam(it, clampPredp(state.z_pred.p, it)), it.contParams.newtonOptions; normN = normC, callback = callback, kwargs...)
	return BorderedArray(res[1], state.z_pred.p), res[2:end]...
end

function getTangent!(iter::AbstractContinuationIterable,
					state::AbstractContinuationState,
					z_new,
					pred::NaturalPred, verbosity)
	(verbosity > 0) && println("Predictor: ", pred)
	# we do nothing here, the predictor will just copy z_old into z_pred
end
####################################################################################################
"""
	Secant tangent predictor
"""
struct SecantPred <: AbstractSecantPredictor end

# τ is the tangent prediction.
# This function is used for initialisation in iterateFromTwoPoints
function getTangent!(τ::M, z_new::M, z_old::M, it::AbstractContinuationIterable, ds, θ, pred::SecantPred, verbosity) where {T, vectype, M <: BorderedArray{vectype, T}}
	(verbosity > 0) && println("Predictor: ", pred)
	# secant predictor: tau = z_new - z_old; tau *= sign(ds) / normtheta(tau)
	copyto!(τ, z_new)
	minus!(τ, z_old)
	α = sign(ds) / it.dotθ(τ, θ)
	rmul!(τ, α)
end

getTangent!(iter::AbstractContinuationIterable, state::AbstractContinuationState, z_new, pred::SecantPred, verbosity) = getTangent!(state.τ, z_new, getSolution(state), iter, state.ds, state.θ, pred, verbosity)
####################################################################################################
"""
	Bordered Tangent predictor
"""
struct BorderedPred <: AbstractTangentPredictor end

# tangent computation using Bordered system
# τ is the tangent prediction found by solving
# ┌                           ┐┌  ┐   ┌   ┐
# │      J            dFdl    ││τu│ = │ 0 │
# │  θ/N ⋅ τ.u     (1-θ)⋅τ.p  ││τp│   │ 1 │
# └                           ┘└  ┘   └   ┘
# it is updated inplace
function getTangent!(it::AbstractContinuationIterable, state::AbstractContinuationState, z_new::BorderedArray{vectype, T}, pred::BorderedPred, verbosity) where {T, vectype}
	(verbosity > 0) && println("Predictor: Bordered")
	ϵ = it.contParams.finDiffEps
	τ = state.τ
	θ = state.θ
	# dFdl = (F(z_new.u, z_new.p + ϵ) - F(z_new.u, z_new.p)) / ϵ
	dFdl = it.F(z_new.u, setParam(it, z_new.p + ϵ))
	minus!(dFdl, it.F(z_new.u, setParam(it, z_new.p)))
	rmul!(dFdl, 1/ϵ)

	# compute jacobian
	J = it.J(z_new.u, setParam(it, z_new.p))

	# extract tangent as solution of the above bordered linear system
	τu, τp, flag, itl = it.linearAlgo( it, state,
										J, dFdl,
										0*z_new.u, T(1)) # Right-hand side

	~flag && @warn "Linear solver failed to converge in tangent computation with type ::BorderedPred"

	# we scale τ in order to have ||τ||_θ = 1 and sign <τ, τold> = 1
	α = T(1) / sqrt(it.dotθ(τu, τu, τp, τp, θ))
	α *= sign(it.dotθ(τ.u, τu, τ.p, τp, θ))

	copyto!(τ.u, τu)
	τ.p = τp
	rmul!(τ, α)
end
####################################################################################################
"""
	Multiple Tangent predictor

$(TYPEDFIELDS)

# Constructor(s)

	MultiplePred(pred, x0, α, n)

	MultiplePred(x0, α, n)

- `α` damping in Newton iterations, 0 < α < 1.
- `n` number of predictors
- `x0` example of vector solution to be stored
"""
@with_kw mutable struct MultiplePred{T <: Real, Tvec, Talgo} <: AbstractTangentPredictor
	"Tangent algorithm used"
	tangentalgo::Talgo

	"Save the current tangent"
	τ::Tvec

	"Damping factor"
	α::T

	"Number of predictors"
	nb::Int64

	"Index of the largest converged predictor"
	currentind::Int64 = 0

	"Index for lookup in residual history"
	pmimax::Int64 = 1

	"Maximum index for lookup in residual history"
	imax::Int64 = 4

	"Factor to increase ds upon successful step"
	dsfact::T = 1.5
end
MultiplePred(pred::AbstractTangentPredictor, x0, α::T, nb) where T = MultiplePred(tangentalgo = pred, τ = BorderedArray(x0, T(0)), α = α, nb = nb)
MultiplePred(x0, α, nb) = MultiplePred(SecantPred(), x0, α, nb)
Base.empty!(mpd::MultiplePred) = (mpd.currentind = 1; mpd.pmimax = 1)

# callback for newton
function (mpred::MultiplePred)(state; kwargs...)
	resHist = get(state, :resHist, nothing)
	iteration = get(state, :iteration, 0)
	if mpred.currentind > 1
		return iteration - mpred.pmimax > 0 ? resHist[end] <= mpred.α * resHist[end-mpred.pmimax] : true
	end
	return true
end

function getTangent!(it::AbstractContinuationIterable, state::AbstractContinuationState, z_new::M, pred::MultiplePred{T, M, Talgo}, verbosity) where {T, vectype, M <: BorderedArray{vectype, T}, Talgo}
	(verbosity > 0) && print("Predictor: MultiplePred\n--")
	getTangent!(it, state, z_new, pred.tangentalgo, verbosity)
	# record the tangent for later use
	copyto!(pred.τ, state.τ)
end

function getPredictor!(it::AbstractContinuationIterable, state::AbstractContinuationState, pred::MultiplePred, nrm = false) where {T, vectype, M <: BorderedArray{vectype, T}}
	# we do nothing!
	# empty!(algo)
	return nothing
end

function corrector(it, z_old::M, tau::M, z_pred::M, ds, θ,
		mpred::MultiplePred, linearalgo = MatrixFreeBLS(); normC = norm,
		callback = cbDefault, kwargs...) where {T, vectype, M <: BorderedArray{vectype, T}}
	verbose = it.verbosity
	(verbose > 1) && printstyled(color=:magenta, "──"^35*"\n   ┌─MultiplePred tangent predictor\n")
	# we combine the callbacks for the newton iterations
	cb = (state; k...) -> callback(state; k...) & mpred(state; k...)
	# note that z_pred already contains ds * τ, hence ii=0 corresponds to this case
	for ii in mpred.nb:-1:1
		(verbose > 1) && printstyled(color=:magenta, "   ├─ i = $ii, s(i) = $(ii*ds), converged = [")
		mpred.currentind = ii # record the current index
		zpred = _copy(z_pred)
		axpy!(ii * ds, mpred.τ, zpred)
		# we restore the original callback if it reaches the usual case ii == 0
		zold, res, flag, itnewton, itlinear = corrector(it, z_old, tau, zpred, ds, θ,
				mpred.tangentalgo, linearalgo; normC = normC, callback = cb, kwargs...)
		if verbose > 1
			if flag
				printstyled("YES", color=:green)
			else
				printstyled(" NO", color=:red)
			end
			printstyled("]\n", color=:magenta)
		end
		if flag || ii == 1 # for i==1, we return the result anyway
			return zold, res, flag, itnewton, itlinear
		end
	end
	return zold, res, flag, itnewton, itlinear
end

function stepSizeControl(ds, θ, contparams::ContinuationPar, converged::Bool, it_newton_number::Int, tau::M, mpd::MultiplePred, verbosity) where {T, vectype, M <: BorderedArray{vectype, T}}
	if converged == false
		dsnew = ds
		if abs(ds) < (1 + mpd.nb) * contparams.dsmin
			if mpd.pmimax < mpd.imax
				(verbosity > 0) && printstyled("--> Increase pmimax\n", color=:red)
				mpd.pmimax += 1
			else
				(verbosity > 0) && @error "Failure to converge with given tolerances"
				# we stop the continuation
				return ds, θ, true
			end
		else
			@error "--> Decrease ds"
			dsnew = ds / (1 + mpd.nb)
			(verbosity > 0) && printstyled("Halving continuation step, ds = $(dsnew)\n", color=:red)
		end
	else # the newton correction has converged
		dsnew = ds
		if mpd.currentind == mpd.nb && abs(ds) * mpd.dsfact <= contparams.dsmax
			(verbosity > 0) && @show dsnew
			# println("--> Increase ds")
			dsnew = ds *  mpd.dsfact
		end
	end

	dsnew = clampDs(dsnew, contparams)

	return dsnew, θ, false
end
####################################################################################################
"""
	Polynomial Tangent predictor

$(TYPEDFIELDS)

# Constructor(s)

	PolynomialPred(pred, n, k, v0)

	PolynomialPred(n, k, v0)

- `n` order of the polynomial
- `k` length of the last solutions vector used for the polynomial fit
- `v0` example of solution to be stored. It is only used to get the `eltype` of the tangent!!
"""
mutable struct PolynomialPred{T <: Real, Tvec, Talgo} <: AbstractTangentPredictor
	"Order of the polynomial"
	n::Int64

	"Length of the last solutions vector used for the polynomial fit"
	k::Int64

	"Matrix for the interpolation"
	A::Matrix{T}

	"Algo for tangent when polynomial predictor is not possible"
	tangentalgo::Talgo

	"Vector of solutions"
	solutions::CircularBuffer{Tvec}

	"Vector of parameters"
	parameters::CircularBuffer{T}

	"Vector of arclengths"
	arclengths::CircularBuffer{T}

	"Coefficients for the polynomials for the solution"
	coeffsSol::Vector{Tvec}

	"Coefficients for the polynomials for the parameter"
	coeffsPar::Vector{T}

	"Update the predictor by adding the last point (x, p)? This can be disabled in order to just use the polynomial prediction. It is useful when the predictor is called mutiple times during bifurcation detection using bisection."
	update::Bool
end

function PolynomialPred(pred, n, k, v0)
	@assert n<k "k must be larger than the degree of the polynomial"
	PolynomialPred(n,k,zeros(eltype(v0), k, n+1), pred,
		CircularBuffer{typeof(v0)}(k),  # solutions
		CircularBuffer{eltype(v0)}(k),  # parameters
		CircularBuffer{eltype(v0)}(k),  # arclengths
		Vector{typeof(v0)}(undef, n+1), # coeffsSol
		Vector{eltype(v0)}(undef, n+1), # coeffsPar
		true)
end
PolynomialPred(n, k, v0) = PolynomialPred(SecantPred(), n, k, v0)

isready(ppd::PolynomialPred) = length(ppd.solutions) >= ppd.k

function Base.empty!(ppd::PolynomialPred)
	empty!(ppd.solutions); empty!(ppd.parameters); empty!(ppd.arclengths);
end

function getStats(polypred::PolynomialPred)
	Sbar = sum(polypred.arclengths) / length(polypred.arclengths)
	σ = sqrt(sum(x->(x-Sbar)^2, polypred.arclengths ) / length(polypred.arclengths))
	# return 0,1
	return Sbar, σ
end

function (polypred::PolynomialPred)(ds::T) where T
	sbar, σ = getStats(polypred)
	s = polypred.arclengths[end] + ds
	snorm = (s-sbar)/σ
	# vector of powers of snorm
	S = Vector{T}(undef, polypred.n+1); S[1] = T(1)
	for jj = 1:polypred.n; S[jj+1] = S[jj] * snorm; end
	p = sum(S .* polypred.coeffsPar)
	x = sum(S .* polypred.coeffsSol)
	return x, p
end

function updatePred!(polypred::PolynomialPred)
	Sbar, σ = getStats(polypred)
	# re-scale the previous arclengths so that the Vandermond matrix is well conditioned
	Ss = (polypred.arclengths .- Sbar) ./ σ
	# construction of the Vandermond Matrix
	polypred.A[:, 1] .= 1
	for jj in 1:polypred.n; polypred.A[:, jj+1] .= polypred.A[:, jj] .* Ss; end
	# invert linear system for least square fitting
	B = (polypred.A' * polypred.A) \ polypred.A'
	mul!(polypred.coeffsSol, B, polypred.solutions)
	mul!(polypred.coeffsPar, B, polypred.parameters)
	return true
end

function getTangent!(it::AbstractContinuationIterable, state::AbstractContinuationState, z_new::M, polypred::PolynomialPred, verbosity) where {T, vectype, M <: BorderedArray{vectype, T}}
	(verbosity > 0) && println("Predictor: PolynomialPred")
	ds = state.ds
	# do we update the predictor with last converged point?
	if polypred.update
		if length(polypred.arclengths) == 0
			push!(polypred.arclengths, ds)
		else
			push!(polypred.arclengths, polypred.arclengths[end]+ds)
		end
		push!(polypred.solutions, z_new.u)
		push!(polypred.parameters, z_new.p)
	end

	if ~isready(polypred) || ~polypred.update
		return getTangent!(it, state, z_new, polypred.tangentalgo, verbosity)
	else
		return polypred.update ? updatePred!(polypred) : true
	end
end

function getPredictor!(z_pred::M, z_old::M, tau::M, ds, polypred::PolynomialPred, nrm = false) where {T, vectype, M <: BorderedArray{vectype, T}}
	if ~isready(polypred)
		return getPredictor!(z_pred, z_old, tau, ds, polypred.tangentalgo, nrm)
	else
		x, p = polypred(ds)
		copyto!(z_pred.u, x)
		z_pred.p = p
		return true
	end
end
####################################################################################################
function arcLengthScaling(θ, contparams, tau::M, verbosity) where {M <: BorderedArray}
	# the arclength scaling algorithm is based on Salinger, Andrew G, Nawaf M Bou-Rabee,
	# Elizabeth A Burroughs, Roger P Pawlowski, Richard B Lehoucq, Louis Romero, and Edward D
	# Wilkes. “LOCA 1.0 Library of Continuation Algorithms: Theory and Implementation Manual,
	# ” March 1, 2002. https://doi.org/10.2172/800778.
	thetanew = θ
	g = abs(tau.p * θ)
	(verbosity > 0) && print("Theta changes from $(θ) to ")
	if (g > contparams.gMax)
		thetanew = contparams.gGoal / tau.p * sqrt( abs(1.0 - g^2) / abs(1.0 - tau.p^2) )
		if (thetanew < contparams.thetaMin)
		  thetanew = contparams.thetaMin;
		end
	end
	(verbosity > 0) && print("$(thetanew)\n")
	return thetanew
end
####################################################################################################
function stepSizeControl(ds, θ, contparams::ContinuationPar, converged::Bool, it_newton_number::Int, tau::M, pred::AbstractTangentPredictor, verbosity) where {T, vectype, M<:BorderedArray{vectype, T}}
	if converged == false
		if  abs(ds) <= contparams.dsmin
			@error "Failure to converge with given tolerances."
			# we stop the continuation
			return ds, θ, true
		end
		dsnew = sign(ds) * max(abs(ds) / 2, contparams.dsmin);
		(verbosity > 0) && printstyled("Halving continuation step, ds=$(dsnew)\n", color=:red)
	else
		# control to have the same number of Newton iterations
		Nmax = contparams.newtonOptions.maxIter
		factor = (Nmax - it_newton_number) / Nmax
		dsnew = ds * (1 + contparams.a * factor^2)
	end

	dsnew = clampDs(dsnew, contparams)

	thetanew = contparams.doArcLengthScaling ? arcLengthScaling(θ, contparams, tau, verbosity) : θ
	# we do not stop the continuation
	return dsnew, thetanew, false
end
####################################################################################################
"""
This is the classical Newton-Krylov solver used to solve `F(x, p) = 0` together
with the scalar condition `n(x, p) ≡ θ ⋅ <x - x0, τx> + (1-θ) ⋅ (p - p0) * τp - n0 = 0`. This makes a problem of dimension N + 1.

Here, we specify `p` as a subfield of `par` with the `paramLens::Lens`

# Arguments
- `(x, par) -> F(x, par)` where `par` is a set of parameters like `(a=1.0, b=1)`
- `(x, par) -> Jh(x, par)` the jacobian Jh = ∂xF
"""
function newtonPALC(iter::AbstractContinuationIterable, state::AbstractContinuationState;
					# linearbdalgo = BorderingBLS(DefaultLS()),
					normN = norm,
					callback = cbDefault, kwargs...)

	F = iter.F; Jh = iter.J
	par = iter.par; paramlens = iter.lens
	contparams = iter.contParams
	dotθ = iter.dotθ
	T = eltype(iter)

	z0 = getSolution(state)
	τ0 = state.τ
	z_pred = state.z_pred
	ds = state.ds; θ = state.θ

	@unpack tol, maxIter, verbose, α, αmin, linesearch = contparams.newtonOptions
	@unpack finDiffEps, pMin, pMax = contparams
	linsolver = iter.linearAlgo

	# we record the damping parameter
	α0 = α

	# N = θ⋅dot(x - z0.u, τ0.u) + (1 - θ)⋅(p - z0.p)⋅τ0.p - ds
	N(u, _p) = arcLengthEq(dotθ, minus(u, z0.u), _p - z0.p, τ0.u, τ0.p, θ, ds)
	normAC(resf, resn) = max(normN(resf), abs(resn))

	# Initialise iterations
	x = _copy(z_pred.u)
	p = z_pred.p
	x_pred = _copy(x)

	res_f = F(x, set(par, paramlens, p));  res_n = N(x, p)

	dX = _copy(res_f)
	dp = T(0)
	up = T(0)
	# dFdp = (F(x, p + finDiffEps) - res_f) / finDiffEps
	dFdp = _copy(F(x, set(par, paramlens, p + finDiffEps)))
	minus!(dFdp, res_f)						# dFdp = dFdp - res_f
	rmul!(dFdp, T(1) / finDiffEps)

	res     = normAC(res_f, res_n)
	resHist = [res]
	it = 0
	itlineartot = 0

	verbose && displayIteration(it, res)
	line_step = true

	compute = callback((;x, res_f, res, it, contparams, p, resHist, options = (;linsolver)); fromNewton = false, kwargs...)

	while (res > tol) && (it < maxIter) && line_step && compute
		# dFdp = (F(x, p + ϵ) - F(x, p)) / ϵ)
		copyto!(dFdp, F(x, set(par, paramlens, p + finDiffEps)))
			minus!(dFdp, res_f); rmul!(dFdp, T(1) / finDiffEps)

		# compute jacobian
		J = Jh(x, set(par, paramlens, p))
		# solve linear system
		# ┌            ┐┌  ┐   ┌     ┐
		# │ J     dFdp ││u │ = │res_f│
		# │ τ0.u  τ0.p ││up│   │res_n│
		# └            ┘└  ┘   └     ┘
		u, up, flag, itlinear = linsolver(iter, state, J, dFdp, res_f, res_n)
		itlineartot += sum(itlinear)

		if linesearch
			line_step = false
			while !line_step && (α > αmin)
				# x_pred = x - α * u
				copyto!(x_pred, x); axpy!(-α, u, x_pred)

				p_pred = p - α * up
				copyto!(res_f, F(x_pred, set(par, paramlens, p_pred)))

				res_n  = N(x_pred, p_pred)
				res = normAC(res_f, res_n)

				if res < resHist[end]
					if (res < resHist[end] / 4) && (α < 1)
						α *= 2
					end
					line_step = true
					copyto!(x, x_pred)

					# p = p_pred
					p  = clamp(p_pred, pMin, pMax)
				else
					α /= 2
				end
			end
			# we put back the initial value
			α = α0
		else
			minus!(x, u)
			p = clamp(p - up, pMin, pMax)

			copyto!(res_f, F(x, set(par, paramlens, p)))

			res_n  = N(x, p); res = normAC(res_f, res_n)
		end

		push!(resHist, res)
		it += 1

		verbose && displayIteration(it, res, itlinear)

		# shall we break the loop?
		compute = callback((;x, res_f, J, res, it, itlinear, contparams, z0, p, resHist, options = (;linsolver)); fromNewton = false, kwargs...)
	end
	verbose && displayIteration(it, res, 0, true) # display last line of the table
	flag = (resHist[end] < tol) & callback((;x, res_f, res, it, contparams, p, resHist, options = (;linsolver)); fromNewton = false, kwargs...)
	return BorderedArray(x, p), resHist, flag, it, itlineartot
end

# conveniency for use in continuation
####################################################################################################
